//
//  CPUUsageMonitor.m
//  CPU Usage
//
//  Created by Peter Hosey on 2006-06-21.
//  Copyright 2006 Peter Hosey. All rights reserved.
//

#import "CPUUsageMonitor.h"

#import "CPUUsageView.h"
#import "NSString+Percentage.h"

@implementation CPUUsageMonitor

+ (void)initialize {
	[self exposeBinding:@"backgroundColor"];
	[self exposeBinding:@"cellWidth"];
	[self exposeBinding:@"cellHeight"];
}

- init {
	if((self = [super init])) {
		NSNumber *thirtyTwo = [NSNumber numberWithFloat:32.0f];
		NSDictionary *defaultPrefs = [NSDictionary dictionaryWithObjectsAndKeys:
			[NSArchiver archivedDataWithRootObject:[NSColor redColor]], @"Background color",
			thirtyTwo, @"Cell width",
			thirtyTwo, @"Cell height",
			[NSNumber numberWithInt:CPUUsageOrientationVertical], @"Window orientation",
			nil];

		[[NSUserDefaults standardUserDefaults] registerDefaults:defaultPrefs];
	}
	return self;
}

- (void)applicationDidFinishLaunching:(NSNotification *)notification {
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	NSData *data = [defaults dataForKey:@"Background color"];
	NSColor *color = data ? [NSUnarchiver unarchiveObjectWithData:data] : [NSColor redColor];
	[self setBackgroundColor:color];

	[self setCellWidth:[defaults floatForKey:@"Cell width"]];
	[self setCellHeight:[defaults floatForKey:@"Cell height"]];

	NSUserDefaultsController *udc = [NSUserDefaultsController sharedUserDefaultsController];
	NSDictionary *backgroundColorOptions = [NSDictionary dictionaryWithObject:@"NSUnarchiveFromData" forKey:NSValueTransformerNameBindingOption];
	[self bind:@"backgroundColor" toObject:udc withKeyPath:@"values.Background color" options:backgroundColorOptions];
	[self bind:@"cellWidth"       toObject:udc withKeyPath:@"values.Cell width"       options:nil];
	[self bind:@"cellHeight"      toObject:udc withKeyPath:@"values.Cell height"      options:nil];

	orientation = [defaults integerForKey:@"Window orientation"];
	BOOL isHorizontal = (orientation == CPUUsageOrientationHorizontal);

	int horizontalStates[2] = { NSOffState, NSOnState };
	[horizontalButton setState:horizontalStates [isHorizontal]];
	[verticalButton   setState:horizontalStates[!isHorizontal]];

	natural_t numProcessors = 0;
	processor_info_array_t processorInfo_nobodyCares;
	mach_msg_type_number_t numProcessorInfo_nobodyCares;
	kern_return_t err = host_processor_info(mach_host_self(), PROCESSOR_CPU_LOAD_INFO, &numProcessors, &processorInfo_nobodyCares, &numProcessorInfo_nobodyCares);
	if(err != KERN_SUCCESS)
		numProcessors = 1; //XXX Should probably error out instead…

	if(numProcessors == 1 && NO) {
		updateTimer = [[NSTimer scheduledTimerWithTimeInterval:0.5
														target:self
													  selector:@selector(updateCPUUsage:)
													  userInfo:nil
													   repeats:YES] retain];
	} else {
		CPUUsageLock = [[NSLock alloc] init];
		CPUUsage = NSZoneMalloc([self zone], numProcessors * sizeof(float));

		updateTimer = [[NSTimer scheduledTimerWithTimeInterval:0.5
														target:self
													  selector:@selector(updateCPUUsageButNotViews:)
													  userInfo:nil
													   repeats:YES] retain];
		[updateTimer fire];

		//Launch the threaded timers that sweep our CPU usage array looking for views.
		for(unsigned i = 0U; i < numProcessors; ++i) {
			[NSThread detachNewThreadSelector:@selector(threadedLaunchTimer:)
									 toTarget:self
								   withObject:[NSNumber numberWithUnsignedInt:i]];
		}
	}

	NSRect frame = {
		.origin = NSZeroPoint,
		.size = {
			cellWidth  * (((numProcessors - 1) *  isHorizontal) + 1),
			cellHeight * (((numProcessors - 1) * !isHorizontal) + 1)
		}
	};
	window = [[NSWindow alloc] initWithContentRect:frame
										 styleMask:NSBorderlessWindowMask
										   backing:NSBackingStoreBuffered
											 defer:YES];
	[window setTitle:@"CPU Usage"];
	[window setOpaque:NO];
	[window setAcceptsMouseMovedEvents:NO];
	[window setMovableByWindowBackground:YES];
	[window setLevel:NSStatusWindowLevel];
	[window setHasShadow:NO];
	[window setIgnoresMouseEvents:YES];
	[window setReleasedWhenClosed:YES];
	[window setResizeIncrements:(NSSize){ numProcessors * isHorizontal, numProcessors * !isHorizontal }];

	[window setFrameAutosaveName:@"CPU usage window"];
	if(![window setFrameUsingName:@"CPU usage window"])
		[window center];
	[preferencesPanel center];

	//Add CPU usage views.

	//If vertical, work from the top down.
	frame.origin.y = cellHeight * ((numProcessors - 1) * isHorizontal);

	usageViews = [[NSMutableArray alloc] initWithCapacity:numProcessors];
	NSView *contentView = [window contentView];
	unsigned CPUNum = 0U;
	for(unsigned i = 0U; i < numProcessors; ++i) {
		frame.origin.x += cellWidth  * (i *  isHorizontal);
		frame.origin.y -= cellHeight * (i * !isHorizontal);

		CPUUsageView *usageView = [[CPUUsageView alloc] initWithFrame:frame];
		unsigned mask = NSViewWidthSizable | NSViewHeightSizable
			| (NSViewMinXMargin * (i  > 0U)                  *  isHorizontal)
			| (NSViewMaxXMargin * (i < (numProcessors - 1U)) *  isHorizontal)
			| (NSViewMinYMargin * (i  > 0U)                  * !isHorizontal)
			| (NSViewMaxYMargin * (i < (numProcessors - 1U)) * !isHorizontal)
			;
		[usageView setAutoresizingMask:mask];
		[usageView setMenu:contextualMenu];

		[usageView setDrawsFrame:NO];
		if(numProcessors > 1)
			[usageView setCPUNumber:++CPUNum];

		[self addObserver:usageView
			   forKeyPath:@"backgroundColor"
				  options:NSKeyValueObservingOptionNew
				  context:NULL];

		[contentView addSubview:usageView];
		[usageViews addObject:usageView];

		[usageView release];
	}

	[window setBackgroundColor:[NSColor clearColor]];

	[window orderFront:nil];
}

- (void)applicationWillTerminate:(NSNotification *)notification {
	//First, fade out the window.
	NSDictionary *animDict = [NSDictionary dictionaryWithObjectsAndKeys:
		window, NSViewAnimationTargetKey,
		NSViewAnimationFadeOutEffect, NSViewAnimationEffectKey,
		nil];
	NSViewAnimation *anim = [[NSViewAnimation alloc] initWithViewAnimations:[NSArray arrayWithObject:animDict]];
	[anim setDuration:0.5];
	[anim setAnimationBlockingMode:NSAnimationNonblockingThreaded];
	[anim startAnimation];

	//Tell the threads to stop, then wait for them to comply.
	threadsShouldExit = YES;
	[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];

	[updateTimer invalidate];
	[updateTimer release];

	[usageViews release];

	[backgroundColor release];

	[window close];
}

- (void)applicationDidBecomeActive:(NSNotification *)notification {
	NSEnumerator *usageViewsEnum = [usageViews objectEnumerator];
	CPUUsageView *usageView;
	while((usageView = [usageViewsEnum nextObject])) {
		[usageView setDrawsFrame:YES];
		[usageView setNeedsDisplay:YES];
	}

	[window setIgnoresMouseEvents:NO];
}
- (void)applicationWillResignActive:(NSNotification *)notification {
	NSEnumerator *usageViewsEnum = [usageViews objectEnumerator];
	CPUUsageView *usageView;
	while((usageView = [usageViewsEnum nextObject])) {
		[usageView setDrawsFrame:NO];
		[usageView setNeedsDisplay:YES];
	}
	
	[window setIgnoresMouseEvents:YES];
}

#pragma mark Timer

- (void)threadedUpdateCPUUsageView:(NSTimer *)timer {
	unsigned i = [[timer userInfo] unsignedIntValue];

	CPUUsageView *view = [usageViews objectAtIndex:i];

	[CPUUsageLock lock];
	[view setCPUUsage:CPUUsage[i]];
	[CPUUsageLock unlock];

	[view display];
}
- (void)threadedLaunchTimer:(NSNumber *)CPUIndexNum {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];

	NSTimer *timer = [NSTimer timerWithTimeInterval:0.5
											 target:self
										   selector:@selector(threadedUpdateCPUUsageView:)
										   userInfo:CPUIndexNum //userInfo
											repeats:YES];
	[timer setFireDate:[NSDate dateWithTimeIntervalSinceNow:0.25]];

	[runLoop addTimer:timer forMode:NSDefaultRunLoopMode];
	[runLoop run];

	NSLog(@"Thread %p exiting", [NSThread currentThread]);
	[pool release];
}

//Main thread.
- (void)updateCPUUsageButNotViews:(NSTimer *)timer {
	natural_t numProcessors = 0U;
	processor_info_array_t processorInfo;
	mach_msg_type_number_t numProcessorInfo;

	kern_return_t err = host_processor_info(mach_host_self(), PROCESSOR_CPU_LOAD_INFO, &numProcessors, &processorInfo, &numProcessorInfo);
	if(err == KERN_SUCCESS) {
		vm_map_t target_task = mach_task_self();

		[CPUUsageLock lock];

		for(unsigned i = 0U; i < numProcessors; ++i) {
			//We only want the last $REFRESH_TIME seconds' worth of data, not all time.
			float inUse, total;
			if(lastProcessorInfo) {
				inUse = (
				  (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER]   - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER])
				+ (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM] - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM])
				+ (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE]   - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE])
				);
				total = inUse + (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE] - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE]);

				size_t lastProcessorInfoSize = sizeof(integer_t) * numLastProcessorInfo;
				vm_deallocate(target_task, (vm_address_t)lastProcessorInfo, lastProcessorInfoSize);
			} else {
				inUse = processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER] + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM] + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE];
				total = inUse + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE];
			}

//			float fraction = inUse / total;
//			NSLog(@"user: %i; system: %i; nice: %i; idle: %i; total: %hf; fraction: %hf", processorInfo[CPU_STATE_USER], processorInfo[CPU_STATE_SYSTEM], processorInfo[CPU_STATE_NICE], processorInfo[CPU_STATE_IDLE], total, fraction);

			CPUUsage[i] = inUse / total;
		}

		[CPUUsageLock unlock];

		lastProcessorInfo = processorInfo;
		numLastProcessorInfo = numProcessorInfo;
	} else {
		NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:
			NSLocalizedString(@"Could not get CPU load information", /*comment*/ nil), NSLocalizedDescriptionKey,
			NSLocalizedString(@"Do you have a CPU in your computer? (Just kidding. I have no idea how to get a useful string from a kernel error code. --The Author)", /*comment*/ nil), NSLocalizedRecoverySuggestionErrorKey,
			nil];
		[NSApp presentError:[NSError errorWithDomain:NSMachErrorDomain code:err userInfo:dict]];
		[NSApp terminate:nil];
	}
}

#if 0
//Single-threaded version.
- (void)updateCPUUsage:(NSTimer *)timer {
	natural_t numProcessors = 0U;
	processor_info_array_t processorInfo;
	mach_msg_type_number_t numProcessorInfo;

	kern_return_t err = host_processor_info(mach_host_self(), PROCESSOR_CPU_LOAD_INFO, &numProcessors, &processorInfo, &numProcessorInfo);
	if(err == KERN_SUCCESS) {
		for(unsigned i = 0U; i < numProcessors; ++i) {
			//We only want the last $REFRESH_TIME seconds' worth of data, not all time.
			float inUse, total;
			if(lastProcessorInfo) {
				inUse = (
				  (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER]   - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER])
				+ (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM] - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM])
				+ (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE]   - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE])
				);
				total = inUse + (processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE] - lastProcessorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE]);
			} else {
				inUse = processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_USER] + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_SYSTEM] + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_NICE];
				total = inUse + processorInfo[(CPU_STATE_MAX * i) + CPU_STATE_IDLE];
			}

			float fraction = inUse / total;
//			NSLog(@"user: %i; system: %i; nice: %i; idle: %i; total: %hf; fraction: %hf", processorInfo[CPU_STATE_USER], processorInfo[CPU_STATE_SYSTEM], processorInfo[CPU_STATE_NICE], processorInfo[CPU_STATE_IDLE], total, fraction);

			CPUUsageView *usageView = [usageViews objectAtIndex:i];
			[usageView setCPUUsage:fraction];
			[usageView setNeedsDisplay:YES];
		}

		size_t lastProcessorInfoSize = sizeof(integer_t) * numProcessorInfo;
		if((!lastProcessorInfo) || (numProcessorInfo != numLastProcessorInfo))
			lastProcessorInfo = realloc(lastProcessorInfo, lastProcessorInfoSize);
		memcpy(lastProcessorInfo, processorInfo, lastProcessorInfoSize);
		numLastProcessorInfo = numProcessorInfo;

		vm_deallocate(mach_task_self(), (vm_address_t)processorInfo, lastProcessorInfoSize);
	} else {
		NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:
			NSLocalizedString(@"Could not get CPU load information", /*comment*/ nil), NSLocalizedDescriptionKey,
			NSLocalizedString(@"Do you have a CPU in your computer? (Just kidding. I have no idea how to get a useful string from a kernel error code. --The Author)", /*comment*/ nil), NSLocalizedRecoverySuggestionErrorKey,
			nil];
		[NSApp presentError:[NSError errorWithDomain:NSMachErrorDomain code:err userInfo:dict]];
		[NSApp terminate:nil];
	}
}
#endif

#pragma mark Accessors

- (NSColor *)backgroundColor {
	return backgroundColor;
}
- (void)setBackgroundColor:(NSColor *)newBackgroundColor {
	if(backgroundColor != newBackgroundColor) {
		[backgroundColor release];
		backgroundColor = [newBackgroundColor retain];
	}
}

- (float)cellWidth {
	return cellWidth;
}
- (void)setCellWidth:(float)newCellWidth {
	cellWidth = newCellWidth;

	if(window) {
		float multiplier = [usageViews count];
		if(multiplier < 0.5f)
			multiplier = 1.0f; //Expect the unexpected.

		NSRect oldFrame = [window frame];
		NSRect newFrame = oldFrame;
		newFrame.size.width = cellWidth * multiplier;
//		[window setFrame:newFrame display:YES animate:YES]; //For some reason, this blocks dragging the window around. Lucky me — the NSViewAnimation is prettier.
		NSDictionary *animDict = [NSDictionary dictionaryWithObjectsAndKeys:
			window, NSViewAnimationTargetKey,
			[NSValue valueWithRect:oldFrame], NSViewAnimationStartFrameKey,
			[NSValue valueWithRect:newFrame], NSViewAnimationEndFrameKey,
			nil];
		NSViewAnimation *anim = [[NSViewAnimation alloc] initWithViewAnimations:[NSArray arrayWithObject:animDict]];
		[anim startAnimation];
		[anim release];
	}
}

- (float)cellHeight {
	return cellHeight;
}
- (void)setCellHeight:(float)newCellHeight {
	cellHeight = newCellHeight;

	if(window) {
		float multiplier = [usageViews count];
		if(multiplier < 0.5f)
			multiplier = 1.0f; //Expect the unexpected.

		NSRect oldFrame = [window frame];
		NSRect newFrame = oldFrame;
		newFrame.size.height = cellHeight * multiplier;
//		[window setFrame:frame display:YES animate:YES]; //For some reason, this blocks dragging the window around. Lucky me — the NSViewAnimation is prettier.
		NSDictionary *animDict = [NSDictionary dictionaryWithObjectsAndKeys:
			window, NSViewAnimationTargetKey,
			[NSValue valueWithRect:oldFrame], NSViewAnimationStartFrameKey,
			[NSValue valueWithRect:newFrame], NSViewAnimationEndFrameKey,
			nil];
		NSViewAnimation *anim = [[NSViewAnimation alloc] initWithViewAnimations:[NSArray arrayWithObject:animDict]];
		[anim startAnimation];
		[anim release];
	}
}

#pragma mark Actions

- (IBAction)setOrientationToVertical:sender {
	if(orientation == CPUUsageOrientationHorizontal) {
		unsigned numProcessors = [usageViews count];

		NSRect windowFrame = [window frame];
		windowFrame.size.width  = cellWidth;
		windowFrame.size.height = cellHeight * numProcessors;
		//Mmmm, I wish I could use a view animation here. The cells would look odd, though.
		[window setFrame:windowFrame display:NO animate:NO];

		NSRect frame = { { 0.0f, cellHeight * (numProcessors - 1U) }, { cellWidth, cellHeight } };

		if(numProcessors == 1U) {
			CPUUsageView *usageView = [usageViews objectAtIndex:0U];
			[usageView setFrame:frame];
			[usageView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
		} else {
			NSEnumerator *usageViewsEnum = [usageViews objectEnumerator];
			CPUUsageView *usageView;
			while((usageView = [usageViewsEnum nextObject])) {
				[usageView setFrame:frame];
				[usageView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
				frame.origin.y -= cellHeight;
			}

			[[usageViews objectAtIndex:0U] setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable | NSViewMaxYMargin];
			[[usageViews objectAtIndex:numProcessors - 1U] setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable | NSViewMinYMargin];
		}

		orientation = CPUUsageOrientationVertical;
		[[[NSUserDefaultsController sharedUserDefaultsController] values] setValue:[NSNumber numberWithInt:orientation] forKey:@"Window orientation"];
	}

	[verticalButton setState:NSOnState];
	[horizontalButton setState:NSOffState];
}
- (IBAction)setOrientationToHorizontal:sender {
	if(orientation == CPUUsageOrientationVertical) {
		unsigned numProcessors = [usageViews count];

		NSRect windowFrame = [window frame];
		windowFrame.size.width  = cellWidth * numProcessors;
		windowFrame.size.height = cellHeight;
		//Mmmm, I wish I could use a view animation here. The cells would look odd, though.
		[window setFrame:windowFrame display:NO animate:NO];

		NSRect frame = { NSZeroPoint, { cellWidth, cellHeight } };

		if(numProcessors == 1U) {
			CPUUsageView *usageView = [usageViews objectAtIndex:0U];
			[usageView setFrame:frame];
			[usageView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
		} else {
			NSEnumerator *usageViewsEnum = [usageViews objectEnumerator];
			CPUUsageView *usageView;
			while((usageView = [usageViewsEnum nextObject])) {
				[usageView setFrame:frame];
				[usageView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
				frame.origin.x += cellWidth;
			}

			[[usageViews objectAtIndex:0U] setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable | NSViewMaxXMargin];
			[[usageViews objectAtIndex:numProcessors - 1U] setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable | NSViewMinXMargin];
		}

		orientation = CPUUsageOrientationHorizontal;
		[[[NSUserDefaultsController sharedUserDefaultsController] values] setValue:[NSNumber numberWithInt:orientation] forKey:@"Window orientation"];
	}

	[verticalButton setState:NSOffState];
	[horizontalButton setState:NSOnState];
}

@end
